using LinearAlgebra
using SparseArrays
using Arpack

function FillOperator(Op::Function,Nx::Int,Ny::Int,Nz::Int)

    Opm = Array{Complex}(undef,0)
    Iv = Array{Int}(undef,0)
    Jv = Array{Int}(undef,0)
    
    s(l,m,n,a) = 2*(Ny+1)*(Nz+1)*l+2*(Nz+1)*m+2*n+a
    for l in 0:Nx    
        for m in 0:Ny
            for n in 0:Nz
                for a in 1:2
                    for l_ in 0:Nx    
                        for m_ in 0:Ny
                            for n_ in 0:Nz
                                for a_ in 1:2
                                    sbra = s(l_,m_,n_,a_)
                                    sket = s(l,m,n,a)
                                    
                                    Opi = Op(l_,m_,n_,a_,l,m,n,a)
                                    if Opi!=0 || sbra==sket
                                        push!(Iv,sbra)
                                        push!(Jv,sket)
                                        push!(Opm,Opi)
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end
    end
    return sparse(Iv,Jv,Opm)
end

"""
     Specialized implementation when nonzero elements are abs(m-m_)+abs(n-n_)+abs(l-l_)<=1
"""
function FillOperatorOptimized(Op::Function,Nx::Int,Ny::Int,Nz::Int)
    ### Need to do special care if Is becomes nondiognal!
    #Hm[:,:] = 0
    Opm = Array{Complex}(undef,0)
    Iv = Array{Int}(undef,0)
    Jv = Array{Int}(undef,0)
    
    s(l,m,n,a) = 2*(Ny+1)*(Nz+1)*l+2*(Nz+1)*m+2*n+a
    for l in 0:Nx    
        for m in 0:Ny
            for n in 0:Nz
                elements = [
                    (l-1,m,n)
                    (l,m,n)
                    (l+1,m,n)
                    (l,m-1,n)
                    (l,m+1,n)
                    (l,m,n-1)
                    (l,m,n+1)
                ]
                
                for (l_,m_,n_) in elements
                    ### A continue routine
                    if (l_<0) || (m_<0) || (n_<0) || (l_>Nx) || (m_>Ny) || (n_>Nz)
                        continue
                    end
                    for a in 1:2
                        for a_ in 1:2
                            sbra = s(l_,m_,n_,a_)
                            sket = s(l,m,n,a)
                            Opi = Op(l_,m_,n_,a_,l,m,n,a)
                            if Hmi!=0
                                push!(Iv,sbra)
                                push!(Jv,sket)
                                push!(Opm,Opi)
                            end
                        end
                    end
                end
            end
        end
    end
    return sparse(Iv,Jv,Opm)
end

### The Hamiltonian

struct ParametersJ
    hw::Tuple{Float64,Float64,Float64}
    I::Tuple{Float64,Float64,Float64}
    phi::Tuple{Float64,Float64,Float64}
end

const sigmax = [0 1;1 0] 
const sigmay = [0 -1im;1im 0]
const sigmaz = [1 0;0 -1] 

δ(m,n) = m==n ? 1 : 0

function H(l_,m_,n_,a_,l,m,n,a,p::ParametersJ)

    hw,I,phi= p.hw,p.I,p.phi

    sigma = (sigmax[a_,a],sigmay[a_,a],sigmaz[a_,a]) 
    bc = (sqrt(l+1)*δ(l_,l+1), sqrt(m+1)*δ(m_,m+1), sqrt(n+1)*δ(n_,n+1))
    ba = (sqrt(l)*δ(l_,l-1),sqrt(m)*δ(m_,m-1),sqrt(n)*δ(n_,n-1))

    E = δ(l_,l)*δ(m_,m)*δ(n_,n)

    s::Complex{Float64} = 0

    s += sum(hw.*(l+1/2,m+1/2,n+1/2).*E.*δ(a_,a))
    s += sum(sigma.*I.*(bc.+ba).*(δ(m_,m)*δ(n_,n),δ(l_,l)*δ(n_,n),δ(l_,l)*δ(m_,m)))
    s += sum(sigma.*I.*phi)*E
    return s
end

### Calculation of energies

function energies(hw,I,phi,N,NEV)
    println("phi = $phi")
    #run(`hostname`)
    BLAS.set_num_threads(1)
    p = ParametersJ(hw,I,phi)

    Hm = FillOperator((l_,m_,n_,a_,l,m,n,a)-> abs(l-l_)+abs(n-n_)+abs(m-m_) > 1 ? 0 : H(l_,m_,n_,a_,l,m,n,a,p),N...)
    solution = eigs(Hm,nev=NEV,tol=1e-18,maxiter=1000,which=:SR)

    # Calculation of spin
    spinx = []
    spiny = []
    spinz = []
    for i in 1:length(solution[1])
        up=solution[2][:,i][1:2:end]
        down=solution[2][:,i][2:2:end]

        spinxi = real(up'*down + down'*up)
        spinyi = real(-1im*up'*down + 1im*down'*up)
        spinzi = up'*up - down'*down

        push!(spinx,spinxi)
        push!(spiny,spinyi)
        push!(spinz,spinzi)
    end

    return Dict(:EVALS=>real(solution[1]),:spinx=>spinx,:spiny=>spiny,:spinz=>spinz)
end
